#!/usr/bin/perl
# Copyright (c) 2008, Rob Bloom
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ROB BLOOM ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL ROB BLOOM BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

use strict;
use PDF::Create;

my %stars;
my %constellations;


my @symbols = ('a','b','g', 'd', 'ep', 'z', 'et', 't', 'i', 'k', 'l', 'm', 'n', 'x',);
my @rainbow = (
	[1,0,0],
	[1,5,0],
	[1,1,0],
	[0,1,0],
	[0,.5,1],
	[0,0,1],
	[.5,0,1],
	[1,0,1]
);

my @starnames;
my @starnums;

my @strings;
open DAT, "<starcon.txt" or die "Cannot open strings file!";
while (<DAT>) {
	if (/^\#\((.*)\)$/) {
		push @strings, $1;
	}
}
close DAT;

my $star_string_count = 133;
my $star_number_base = ( 133 + 29 + 10 + 133 + 56);
my $star_number_count = 14;

for ( ($star_number_base) .. ($star_number_base+$star_number_count-1) ) {
	push @starnums, $strings[$_];
}

for ( 0 .. ($star_string_count-1) ) {
	push @starnames, $strings[$_];
}


open DAT, "<plandata.c" or die "Cannot open planetary data file!";
my $isstardat = 0;

while (<DAT>) {
	if (!$isstardat) {
		if (/^STAR_DESC starmap_array/) {
			$isstardat = 1;
		}
	} else { # $isstardat
		if (/QuasiSpace/) {
			$isstardat = 0;
		}
		if (/\{\{(.*)\}, MAKE_STAR \((.*)\), (.*)\}/) {
			my ($coord, $info, $name) = ($1, $2, $3);
			$coord =~ /(.*), (.*)/;
			my ($x,$y) = ($1,$2);
			$info =~ /(.*), (.*), -1/;
			my ($size, $color) = ($1,$2);
			$name =~ /(.*), (.*), (.*)/;
			my ($interest, $prefix, $postfix) = ($1,$2,$3);

			my $name = toname($prefix,$postfix);
			$stars{$name} = {
				x=>$x,
				y=>$y,
				size=>$size,
				color=>$color,
				interest=>$interest,
				pre=>$prefix,
				post=>$postfix,
			};

			push @{$constellations{$postfix}}, [$x, $y];
		}
	}
}

# Data loaded!  Now to generate interesting stuff.


print "Drawing map!\n";

if (0) {
	foreach (sort keys %stars) {
		my $starref = $stars{$_};
		my %star = %{$starref};
		print "$_:\n";
		foreach (sort keys %star) {
			print "\t$_: ${star{$_}}\n";
		}
	}
}


my $pdf = new PDF::Create(
	filename=>'starmap.pdf',
	Version=>1.2,
	PageMode=>'UseNone',
	Author=>'Hadley <hadley@foxular.net',
	Title=>'Ur-Quan Masters Map of Hyperspace',
	CreationDate=>[ localtime ]
);

my $w = int 11*72;
my $h = int 8.5*72;

my $page = $pdf->new_page(
	MediaBox=>[0,0,$w, $h]
);

my $font = $pdf->font(
	BaseFont=>'Helvetica'
);
my $fontbold = $pdf->font(
	BaseFont=>'Helvetica-Bold'
);

$page->setrgbcolor(0,0,0);
$page->rectangle(0,0,$w,$h);
$page->fill();

$page->setrgbcolor(1,1,1);
$page->stringc($fontbold, 36, $w/2, $h-36, "Ur-Quan Masters");
$page->stringc($fontbold, 36, $w/2, $h-36-36, "Map of HyperSpace");

my $mapstartx = 18;
my $mapstarty = 18;
my $mapsize = 504;

$page->newpath();
$page->set_width(2);
$page->setrgbcolorstroke(.65,.65,.65);
$page->rectangle($mapstartx-1,$mapstarty-1,$mapsize+2,$mapsize+2);
$page->stroke();

# Grid
print "Drawing grid...\n";
for (my $x = 0; $x < 10000; $x += 500) {
	$page->newpath();
	$page->set_width(.25);
	$page->setrgbcolorstroke(.2,.2,.2);
	my ($dx,$dy) = unitomap($x,0);
	$page->moveto($dx,$dy);
	($dx,$dy) = unitomap($x,10000);
	$page->lineto($dx,$dy);
	$page->stroke();
	$page->setrgbcolor(.8,.8,.8);
	$page->stringc($font, 4, $dx, $mapstarty+$mapsize+3, int($x/10));
	$page->stringc($font, 4, $dx, $mapstarty-4-2, int($x/10));

}
for (my $y = 0; $y < 10000; $y += 500) {
	$page->newpath();
	$page->set_width(.25);
	$page->setrgbcolorstroke(.2,.2,.2);
	my ($dx,$dy) = unitomap(0,$y);
	$page->moveto($dx,$dy);
	($dx,$dy) = unitomap(10000,$y);
	$page->lineto($dx,$dy);
	$page->stroke();
	$page->stringr($font, 4, $mapstartx-3, $dy-2, int($y/10));
	$page->stringl($font, 4, $mapstartx+$mapsize+3, $dy-2, int($y/10));
}


print "Drawing list of stars...\n";
my $liststartx = 540;
my $liststarty = 120;
my $listwidth = 234;
my $listheight = 402;

$page->newpath();
$page->setrgbcolorstroke(.65,.65,.65);
$page->set_width(2);
$page->rectangle($liststartx-1,$liststarty-1, $listwidth+2,$listheight+2);
$page->stroke();

$page->stringc($fontbold,16,
	$liststartx + ($listwidth/2),
	$liststarty + $listheight-16, "List of stars");

my $x = $liststartx+2;
my $y = $liststarty + $listheight-24;
foreach (sort constsort keys %stars) {
	my $crd = sprintf("(%.1f, %.1f)",$stars{$_}{'x'}/10, $stars{$_}{'y'}/10);
	$page->string($font,2.8,$x,$y,$crd);
	my $name = $_;
	if ($name =~ / /) {
		$name =~ s/^(.*) (.*)/$2, $1/;
	}
	$page->string($font,2.8,$x+20,$y,$name);
	$y -= 3;
	if ($y < ($liststarty+3)) {
		$x += ($listwidth/4);
		$y = $liststarty + $listheight-24;
	}
}

print "Drawing legend...\n";
my $legstartx = 540;
my $legstarty = 18;
my $legwidth = 234;
my $legheight = 84;
$page->newpath();
$page->setrgbcolorstroke(.65,.65,.65);
$page->set_width(2);
$page->rectangle($legstartx-1,$legstarty-1, $legwidth+2,$legheight+2);
$page->stroke();

$page->stringc($fontbold,16,
	$legstartx + ($legwidth/2),
	$legstarty + $legheight-16, "Legend");

$page->setrgbcolorstroke(1,1,1);
$page->setrgbcolor(1,1,1);
drawcircle($page, $legstartx + 8, $legstarty + $legheight - 32, 1.0/2);
$page->set_width(1.0/5);
$page->stroke();
$page->string($font, 8, $legstartx + 16, $legstarty + $legheight - 34, "Dwarf star");
drawcircle($page, $legstartx + 8, $legstarty + $legheight - 48, 2.0/2);
$page->set_width(2.0/5);
$page->stroke();
$page->string($font, 8, $legstartx + 16, $legstarty + $legheight - 50.5, "Giant star");
drawcircle($page, $legstartx + 8, $legstarty + $legheight - 64, 3.0/2);
$page->set_width(3.0/5);
$page->stroke();
$page->string($font, 8, $legstartx + 16, $legstarty + $legheight - 67, "Super-giant star");

drawcircle($page, $legstartx + 86, $legstarty + $legheight - 48, 2.0/2);
$page->set_width(2.0/5);
$page->stroke();
draw_sparkle($legstartx+86, $legstarty+$legheight-48, 2, 0);
$page->string($font, 8, $legstartx + 94, $legstarty + $legheight - 50.5, "Point of Interest");

drawcircle($page, $legstartx + 86, $legstarty + $legheight - 64, 2.0/2);
$page->set_width(2.0/5);
$page->stroke();
draw_sparkle($legstartx+86, $legstarty+$legheight-64, 2, 1);
$page->string($font, 8, $legstartx + 94, $legstarty + $legheight - 66, "Rainbow Planet");

my $legx = $legstartx + 160;
my $legy = $legstarty + $legheight - 26;
my $i = 0;
foreach (@starnums) {
	$page->string($font, 5, $legx, $legy, $symbols[$i++] . " - " . $_);
	$legy -= 7;
	if ($legy < $legstarty+12) {
		$legy = $legstarty + $legheight - 26;
		$legx += 40;
	}
}

print "Drawing constellations...\n";
foreach (keys %stars) {
	my %star = %{$stars{$_}};
	
	my ($pre,$post) = ($star{'pre'}, $star{'post'});
	if ($pre > 1) {
		my $p = $pre;
		my $par;
		while (!$stars{$par}) {
			$p--;
			$par = toname($p, $post);
		}
		#print "$_ connects to $par\n";
		my %parent = %{$stars{$par}};
		$page->set_width(.25);
		$page->setrgbcolor(.25,.25,.25);
		$page->setrgbcolorstroke(.25,.25,.25);
		my ($x1, $y1) = unitomap($star{'x'}, $star{'y'});
		my ($x2, $y2) = unitomap($parent{'x'}, $parent{'y'});
		$page->line($x1,$y1,$x2,$y2);
	}
}


print "Drawing stars...\n";
foreach (keys %stars) {
	my %star = %{$stars{$_}};

	my ($x, $y) = unitomap($star{'x'}, $star{'y'});
	my $sizes = $star{'size'};
	my $col = $star{'color'};

	$page->newpath();
	my $size = 10;
	if ($sizes eq 'DWARF_STAR') {
		$size = 1.0;
	} elsif ($sizes eq 'GIANT_STAR') {
		$size = 2.0;
	} elsif ($sizes eq 'SUPER_GIANT_STAR') {
		$size = 3.0;
	}

	my ($r,$g,$b);

	if ($col eq 'BLUE_BODY') {
		($r,$g,$b) = (.3,.3,1);
	} elsif ($col eq 'GREEN_BODY') {
		($r,$g,$b) = (.3,1,.3);
	} elsif ($col eq 'ORANGE_BODY') {
		($r,$g,$b) = (1,.5,0);
	} elsif ($col eq 'RED_BODY') {
		($r,$g,$b) = (1,0,0);
	} elsif ($col eq 'WHITE_BODY') {
		($r,$g,$b) = (1,1,1);
	} elsif ($col eq 'YELLOW_BODY') {
		($r,$g,$b) = (1,1,0);
	}


	my $interest = $star{'interest'};
	if ($interest ne '0' && !($interest =~ /MELNORME._DEFINED/)) {
		draw_sparkle($x, $y, $size, $interest eq 'RAINBOW_DEFINED');
	}

	$page->setrgbcolorstroke($r,$g,$b);
	$page->setrgbcolor($r,$g,$b);
	#$page->rectangle($x - ($size/2), $y - ($size/2), $size, $size);
	drawcircle($page, $x, $y, $size/2);
	$page->set_width($size/5);
	#print "$_ at ".int($x).", ".int($y)."... ";
	$page->stroke();

	$page->setrgbcolor(.5,.5,.5);
	if ($star{'pre'}) {
		$page->string($font, 3, $x+($size/3), $y + .25, $symbols[$star{'pre'}-1]);
	}
}
#print "\n";


print "Labelling constellations and repelling labels...\n";
foreach (keys %constellations) {
	my $name = $starnames[$_];
	my @coords = @{$constellations{$_}};
	#print "$name: \n";
	my $amt=0;
	my $avgx=0;
	my $avgy=0;
	foreach (@coords) {
		my ($x, $y) = @{$_};
		#print "\t$x, $y\n";
		$avgx += $x;
		$avgy += $y;
		$amt++;
	}
	$avgx/=$amt;
	$avgy/=$amt;

	($avgx, $avgy) = unitomap($avgx,$avgy);

	my $strw = $page->string_width($fontbold, $name);
	my $x = $avgx;
	my $y = $avgy+4;

	my $valid = 0; # repel labels off of stars until they do not intersect.
	my $moveleft = 200; # Infinite loop prevention
	while (!$valid && $moveleft) {
		$moveleft--;
		if ($x - $strw < $mapstartx + 10 ) {
			$x = $mapstartx + 10 + $strw;
		}
		if ($x + $strw > $mapstartx + $mapsize - 10 ) {
			$x = $mapstartx + $mapsize - $strw - 10;
		}

		if ($y + 2 > $mapstarty + $mapsize - 5) {
			$y = $mapstarty + $mapsize - 5;
		}

		my @col;
		foreach (keys %stars) {
			my %star = %{$stars{$_}};
			my ($sx, $sy) = unitomap($star{'x'}, $star{'y'});
			if (($sx > $x - $strw-4) && 
				($sx < $x + $strw+4) &&
				($sy > $y - 6) &&
				($sy < $y + 6)) {
				push @col, \%star;
			}
		}
		if (@col == 0) {
			$valid = 1;
		} else {
			# Special cases for some labels the repel doesn't work on
			if ($name eq 'Chandrasekhar') {
				$y += 1;
			} elsif ($name eq 'Sol') {
				$x -= 2;
			} elsif ($name eq 'Vega') {
				$x += 3;
				$y -= 2;
			} elsif ($name eq 'Corvi') {
				$x -= 2;
				$y -= 1;
			} elsif ($name eq 'Apodis') {
				$x -= 1;
				$y -= .5;
			} elsif ($name eq 'Draconis') {
				$x += 1;
				$y += .25;
			} elsif ($name eq 'Fornacis') {
				$y += 1;
			} elsif ($name eq 'Eridani') {
				$x -= 1;
				$y -= .5;
			} elsif ($name eq 'Chamaeleonis') {
				$x += 2;
				$y -= .5;
			} elsif ($name eq 'Capricorni') {
				$x -= .5;
				$y += .1;
			} elsif ($name eq 'Sagittae') {
				$y -= 1;
			} elsif ($name eq 'Illuminati') {
				$y -= 1;
			} elsif ($name eq 'Vulpeculae') {
				$x += 1;
			} elsif ($name eq 'Vela') {
				$y -= 1;
			} else {
				my $avgx;
				my $avgy;
				foreach (@col) {
				my ($sx, $sy) = unitomap(${$_}{'x'}, ${$_}{'y'});
					my $dx += $x - $sx;
					my $dy += $y - $sy;
					my $d = sqrt(($dx*$dx)+($dy*$dy));
					$dx /= $d;
					$dy /= $d;
					$avgx += $dx;
					$avgy += $dy;
				}
				$avgx /= ($#col+1);
				$avgy /= ($#col+1);
				my $d = sqrt(($avgx*$avgx)+($avgy*$avgy));
				$x += 2*$avgx/$d;
				$y += 2*$avgy/$d;
				if ($name eq "Chamaeleonis") {
					print "Moving $name $avgx, $avgy\n";
				}
			}
		}
	}

	$page->setrgbcolor(.8,.8,.8);
	#print "\t\tAVG: $avgx, $avgy\n";
	$page->stringc($fontbold,4,$x,$y, $name);
}


print "Done!\n";

$pdf->close();

sub toname {
	my ($pre, $post) = @_;
	if ($pre == 0) {
		return $starnames[$post];
	} else {
		return $starnums[$pre - 1] . " " . $starnames[$post];
	}
}

sub unitomap {
	my ($x, $y) = @_;
	return ($x*$mapsize/9999 + $mapstartx,
			$y*$mapsize/9999 + $mapstarty)
}

sub drawcircle {
	my ($page, $x, $y, $r) = @_;
	#$page->rectangle($x - ($size/2), $y - ($size/2), $size, $size);
	$page->moveto($x-$r, $y);
	my $b = $r * sqrt(2)/2;
	$page->curveto($x-$b, $y+$b, $x-$b, $y+$b, $x, $y+$r);
	$page->curveto($x+$b, $y+$b, $x+$b, $y+$b, $x+$r, $y);
	$page->curveto($x+$b, $y-$b, $x+$b, $y-$b, $x, $y-$r);
	$page->curveto($x-$b, $y-$b, $x-$b, $y-$b, $x-$r, $y);
	$page->closepath();
	return;
}


sub constsort {
	my $apost = $a;
	$apost =~ s/^.* //;
	my $bpost = $b;
	$bpost =~ s/^.* //;
	if (($apost cmp $bpost) != 0) {
		return $apost cmp $bpost;
	}

	return $stars{$a}{'pre'} <=> $stars{$b}{'pre'};
}

sub draw_sparkle {
	my ($x, $y, $size, $rainbow) = @_;
	for (0 .. 7) {
		if ($rainbow) {
			my ($r,$g,$b) = @{$rainbow[$_]};
			$page->setrgbcolorstroke($r,$g,$b);
			$page->set_width($size/3);
		} else {
			$page->setrgbcolorstroke(1,1,1);
			$page->set_width($size/10);
		}
		my $r = $_*3.1415926535897932384626*2/8;
		my $xd = sin($r);
		my $yd = cos($r);
		$page->line($x+($xd*$size*.55),$y+($yd*$size*.55),
					$x+($xd*$size*1.5),$y+($yd*$size*1.5))
	}	
}
